-- == SRB2Randomizer: Base file ==
-- Basic initialization of the main variables.
-- Should always be the first file in the list.

rawset(_G, "randomizer", {})

randomizer.seed = nil -- The meat of any randomizer.
randomizer.seedstring = nil -- If the seed was set using a string value, then we'll display that on the scores HUD, in addition to the number.
randomizer.rv = nil -- seed RNG, advanced by the mapload hooks
randomizer.levellist = {} -- level list for LEVELS.lua

addHook("NetVars", function(net)
	randomizer.seed = net(randomizer.seed)
	randomizer.seedstring = net(randomizer.seedstring)
	randomizer.rv = net(randomizer.rv)
	randomizer.levellist = net(randomizer.levellist)
end)

randomizer.cvars = {}
randomizer.cvarMaxVals = {}
randomizer.cvval = {}
randomizer.cvhash = 0
randomizer.help = {
	server = {},
	client = {}
}

local MAX_CVAR_VALUE = 8

-- Creates a new cvar/command info
function randomizer.newHelpEntry(name, args, tag, detailed, clientside)
	local t = {}

	t.name = name -- name
	t.args = args -- arguments
	t.tag = tag -- single-line help
	t.detailed = detailed -- multi-line help (table)

	if (clientside)
		table.insert(randomizer.help.client, t)
	else
		table.insert(randomizer.help.server, t)
	end
end

randomizer.cvars.debug = CV_RegisterVar({
	name = "randomizer_debug",
	flags = CV_SHOWMODIF,
	PossibleValue = {
		Off = 0,
		On = 1,
		Full = 2,
	},
	defaultvalue = "Off",
})

randomizer.newHelpEntry(
	"debug",
	"<integer/string>",
	"Enables debug mode.",
	{
		"Enables debug mode. The main use is to show debugging prints. Will cause a lot of console spam. Options are...",
		" * \130Off (0):\128 No debugging information. (default option)",
		" * \130On (1):\128 Debugging information will be printed.",
		" * \130Full (2):\128 Same as \"On\", but also shows information about the RNG state.",
	},
	false
)

function randomizer.debugPrint(input)
	if (randomizer.cvars.debug.value != 1)
		return
	end

	print(input)
end

-- Scrambles the value of an inputted seed.
-- Just a very simplistic form of a Tausworthe generator.
function randomizer.scramble(input)
	-- Initial constant XOR, prevents seed 0 from always outputting 0
	-- This value basically means nothing and could be changed to anything
	local output = input ^^ 1064338584

	-- XOR and shift around to get a random-feeling value
	output = $1 ^^ (input >> 13)
	output = $1 ^^ (input << 25)
	output = $1 ^^ (input >> 11)

	-- AND with max value
	output = (abs(output) & UINT32_MAX)

	if (randomizer.cvars.debug.value == 2)
		randomizer.debugPrint(input .. " -> " .. output)
	end

	return output
end

-- Copies a table, used by a few of the other files
function randomizer.simpleTableCopy(input)
	local output = {}
	for i = 1,#input
		output[i] = input[i]
	end
	return output
end


local function randomCommandValue(v)
	if (v == nil)
		v = P_RandomFixed()
	else
		v = abs($1 + P_RandomFixed())
	end

	local scramblecount = 16
	while (scramblecount)
		v = randomizer.scramble(v)
		scramblecount = $1-1
	end

	return v
end

local function initSeed(input)
	local num = tonumber(input)
	local isstring = false

	if (num != nil)
		randomizer.seed = abs(num)
	elseif (type(input) == "string")
		local newseed = 0

		for i = 1,#input
			local b = input:byte(i,i)
			newseed = $1 + (b << i)
		end

		newseed = abs($1)
		randomizer.seed = newseed
		isstring = true
	else
		randomizer.seed = randomCommandValue(randomizer.seed)
	end

	local pr = "SRB2Randomizer seed has been set to \130"..randomizer.seed.."\128."

	if (isstring)
		pr = $1.." (\"\130"..input.."\128\")"
		randomizer.seedstring = input
	else
		randomizer.seedstring = nil
	end

	print(pr)
end

COM_AddCommand("randomizer_setseed", function(blah, arg)
	initSeed(arg)

	for p in players.iterate()
		p.randomizer = {}
	end

	-- See levels.lua
	randomizer.updateLevelList()
end, COM_ADMIN)

randomizer.newHelpEntry(
	"setseed",
	"[<seed value>]",
	"Changes the randomization seed.",
	{
		"This command changes the seed being used. Leave blank to randomly generate one.",
		"Accepts both numbers and strings -- strings will convert into a number.",
		"The seed will be uninitialized until you use this command.",
		"\131NOTICE:\128 If you plan on using a randomly generated seed, it's recommended that you either:",
		" * Play for a little bit before you use the command.",
		" * Use the command several times to advance the RNG.",
		"Otherwise, you may end up with the same seed every time! This is because SRB2's RNG always initalizes to a specific value.",
	},
	false
)

randomizer.cvars.alwaysshowseed = CV_RegisterVar({
	name = "randomizer_alwaysshowseed",
	flags = CV_SHOWMODIF,
	PossibleValue = CV_OnOff,
	defaultvalue = "Off",
})

randomizer.newHelpEntry(
	"alwaysshowseed",
	"<boolean>",
	"Show the current seed on the HUD.",
	{
		"Decide whenever or not to show the current seed/settings hash on the normal gameplay HUD. Options are...",
		" * \130Off (0):\128 Will only be visible on the Rankings/Scores HUD. (default option)",
		" * \130On (1):\128 Will always be visible. Good for ensuring there's no cheating going on in a speedrun race.",
	},
	true
)

local didSeedWarning = false

addHook("MapLoad", function()
	-- This is the "random value" used for all randomizations.
	-- This is changed from the default seed in any MapLoad effects,
	-- so that changing settings gives you different randomization,
	-- and generally make it feel a little more random without impacting seed consistency.
	-- Effects applied mid-game, such as characters, can use this value, but not modify it.
	-- Effects that are supposed to be consistent between different maps should just use the seed directly.
	randomizer.rv = randomizer.seed
	didSeedWarning = false
end)

local printTimer = 0

addHook("ThinkFrame", function()
	if (randomizer.seed == nil)
		if (printTimer >= 2*TICRATE) and not (didSeedWarning)
			-- Because of the way SRB2's RNG is initialized,
			-- providing a random seed when the script loads on the title screen would give you the same seed every time.
			-- That kinda defeats the point of a randomizer, so I put this message here.
			print("\3\131NOTICE:\128 A seed for SRB2Randomizer has not been set. Use the command 'randomizer_setseed' to set one.")
			didSeedWarning = true
		else
			printTimer = $1+1
		end

		return
	end

	for p in players.iterate()
		if (p.randomizer == nil)
			p.randomizer = {}
		end
	end

	randomizer.cvhash = 0

	local i = 0

	for k,v in pairs(randomizer.cvars)
		local cv = v

		if not (cv.flags & CV_NETVAR)
			continue
		end

		randomizer.cvhash = $1 | (cv.value + (MAX_CVAR_VALUE * i))
		i = $1+1
	end
end)

randomizer.newHelpEntry(
	"setseed",
	"[<seed value>]",
	"Changes the randomization seed.",
	{
		"This command changes the seed being used. Leave blank to randomly generate one.",
		"Accepts both numbers and strings -- strings will convert into a number.",
		"The seed will be uninitialized until you use this command.",
		"\131NOTICE:\128 If you plan on using a randomly generated seed, it's recommended that you either:",
		" * Play for a little bit before you use the command.",
		" * Use the command several times to advance the RNG.",
		"Otherwise, you may end up with the same seed every time! This is because SRB2's RNG always initalizes to a specific value.",
	},
	false
)

local function drawSeed(v)
	if (randomizer.seed == nil)
		return
	end

	local seed = randomizer.seed

	if (randomizer.seedstring != nil)
		seed = randomizer.seedstring.." ("..$1..")"
	end

	v.drawString(316, 192, "\130"..seed.."\128 - "..randomizer.cvhash, V_SNAPTOBOTTOM|V_50TRANS, "thin-right")
end

hud.add(function(v, p)
	if (multiplayer) and not (netgame) and (#p != 0)
		return -- no overlapping in SS
	end

	if (randomizer.cvars.alwaysshowseed.value == 1)
		drawSeed(v)
	end
end, "game")

hud.add(function(v)
	if (randomizer.cvars.alwaysshowseed.value == 0)
	or (multiplayer) or (netgame) -- no overlapping in SP
		drawSeed(v)
	end
end, "scores")
