-- == SRB2Randomizer: Prize randomization ==
-- Switches around "major" powerups/collectibles (monitors, tokens, emblems)

randomizer.cvars.monitors = CV_RegisterVar({
	name = "randomizer_monitors",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = {
		Off = 0,
		Replace = 1,
		Shuffle = 2,
	},
	defaultvalue = "Replace",
})

randomizer.newHelpEntry(
	"monitors",
	"<integer/string>",
	"Changes monitor randomization style.",
	{
		"Changes monitor randomization style. Options are...",
		" * \130Off (0):\128 Monitors will not be randomized.",
		" * \130Replace (1):\128 All monitors will be completely replaced with new types. Creates lots of variety, and allows weird/custom monitor types to get mixed in. (default option)",
		" * \130Shuffle (2):\128 Monitors will swap places with each other. Keeps the level's item balance in-tact, if you prefer a tamer experience.",
	},
	false
)

randomizer.cvars.goldmonitors = CV_RegisterVar({
	name = "randomizer_goldmonitors",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = {
		Off = 0,
		On = 1,
		OnlyNormal = 2,
	},
	defaultvalue = "Off",
})

randomizer.newHelpEntry(
	"goldmonitors",
	"<integer/string>",
	"Changes golden monitor randomization style.",
	{
		"Changes golden monitor randomization style. Note that this option requires normal monitors to be randomized as well. Options are...",
		" * \130Off (0):\128 Golden monitors will not be randomized. (default option)",
		" * \130On (1):\128 Golden monitors will be randomized using the same setting as normal monitors.",
		" * \130OnlyNormal (2):\128 Golden monitors will not be randomized, but normal monitors can be changed into golden ones if monitor randomization is set to \"Replace\".",
	},
	false
)


randomizer.cvars.emeralds = CV_RegisterVar({
	name = "randomizer_emeralds",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = {
		Off = 0,
		On = 1,
		Remove = 2,
	},
	defaultvalue = "On",
})

randomizer.newHelpEntry(
	"emeralds",
	"<integer/string>",
	"Changes Emerald Token / Chaos Emerald randomization style.",
	{
		"Changes Emerald Token / Chaos Emerald randomization style. Options are...",
		" * \130Off (0):\128 Emeralds will not be randomized.",
		" * \130On (1):\128 Emeralds will swap places with random monitor spots. Requires extra exploration & map knowledge to get all Chaos Emeralds. (default option)",
		" * \130Remove (2):\128 Remove Emeralds from the game entirely. Their spots will be used to add new monitors.",
		"This option is specifically meant for those map packs that don't add unique Special Stages, but still have Emerald Tokens for some reason. Can also be set if you hate having Super Sonic or hate going to Special Stages.",
	},
	false
)

randomizer.cvars.emblems = CV_RegisterVar({
	name = "randomizer_emblems",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = CV_OnOff,
	defaultvalue = "Off",
})

randomizer.newHelpEntry(
	"emblems",
	"<boolean>",
	"Changes Emblem randomization style.",
	{
		"Changes Emblem randomization style. Options are...",
		" * \130Off (0):\128 Emblems will not be randomized. (default option)",
		" * \130On (1):\128 If emblems are available, then they will swap places with other monitor and emerald spots.",
		"\131NOTICE:\128 For emblems to be available, you will need to be in 1 Player mode, and you will need to add a \"dummy\" save file WAD before playing. Due to these specific circumstances, this is Off by default."
	},
	false
)

local randotokenlist = 0
local lastmap = nil

local function monitorIsSwappable(info)
	if (info.flags & MF_MONITOR)
		if (info.flags & MF_GRENADEBOUNCE)
		and (randomizer.cvars.goldmonitors.value != 1)
			return false
		end

		return true
	end

	return false
end

local function generateMonitorSpotsTable()
	local monitorspots = {}

	for mo in mobjs.iterate()
		if (monitorIsSwappable(mo.info))
			table.insert(monitorspots, mo)
		end
	end

	return monitorspots
end

local function generateMonitorTypesTable()
	local monitortypes = {}

	for i = 0,#mobjinfo-1
		if (mobjinfo[i].flags & MF_MONITOR)
		and (mobjinfo[i].doomednum != -1) -- make sure it's intended to be placeable
		and (G_GametypeHasTeams() or not (i == MT_RING_REDBOX or i == MT_RING_BLUEBOX)) -- CTF boxes
		and ((multiplayer or netgame) or not (i == MT_MIXUP_BOX or i == MT_RECYCLER_BOX)) -- don't spawn recyclers or swaps in singleplayer
			local addamount = 4

			-- Weighted amounts based on type
			if (mobjinfo[i].flags & MF_GRENADEBOUNCE)
				if (randomizer.cvars.goldmonitors.value == 0)
					-- Don't add gold boxes
					continue
				else
					-- Gold boxes should be extremely rare
					addamount = 1
				end
			elseif (i == MT_RING_BOX)
				-- Rings should be a lot more frequent
				addamount = 30
			elseif (i == MT_MIXUP_BOX)
			or (i == MT_RECYCLER_BOX)
			or (i == MT_GRAVITY_BOX)
				-- Weirdo items should be rarer
				addamount = 2
			end

			for j = 1,addamount
				table.insert(monitortypes, i)
			end
		end
	end

	return monitortypes
end

local function replaceMonitors(spots)
	randomizer.debugPrint("Replacing monitors")

	local monitortypes = generateMonitorTypesTable()

	for i = 1,#spots
		local mo = spots[i]

		if (mo.tracer and mo.tracer.valid) -- 1up overlay
			P_RemoveMobj(mo.tracer)
		end

		local newtype = monitortypes[(randomizer.rv % #monitortypes) + 1]
		if (mo.type != newtype)
			mo.type = newtype
			mo.state = mobjinfo[newtype].spawnstate
		end

		--[[
		if (mo.spawnpoint and mo.spawnpoint.valid)
			mo.spawnpoint.type = mobjinfo[mo.type].doomednum
		end
		--]]

		randomizer.rv = randomizer.scramble(randomizer.rv)
	end
end

local function shuffleMonitors(spots)
	randomizer.debugPrint("Shuffling monitors")

	local monitors = randomizer.simpleTableCopy(spots)

	local i = 1
	while (i <= #monitors)
		local mo = monitors[i]

		local swapindex = (randomizer.rv % #monitors) + 1
		local swap = monitors[swapindex]

		if (mo.type != swap.type)
			local temp = mo.type

			mo.type = swap.type
			mo.state = mobjinfo[swap.type].spawnstate

			--[[
			if (mo.spawnpoint and mo.spawnpoint.valid)
				mo.spawnpoint.type = mobjinfo[mo.type].doomednum
			end
			--]]

			swap.type = temp
			swap.state = mobjinfo[temp].spawnstate

			--[[
			if (swap.spawnpoint and swap.spawnpoint.valid)
				swap.spawnpoint.type = mobjinfo[temp].doomednum
			end
			--]]
		end

		table.remove(monitors, swapindex)
		randomizer.rv = randomizer.scramble(randomizer.rv)
		i = $1+1
	end
end

local function shuffleEmeralds()
	randomizer.debugPrint("Shuffling Chaos Emeralds & tokens")

	local monitors = {}
	local tokens = {}

	for mo in mobjs.iterate()
		if not (mo and mo.valid)
			continue
		end

		if (monitorIsSwappable(mo.info))
			if (mo.tracer and mo.tracer.valid) -- 1up overlay
				P_RemoveMobj(mo.tracer)
			end
			table.insert(monitors, mo)
		elseif (mo.type == MT_TOKEN)
		or (mo.type == MT_EMERALD1)
		or (mo.type == MT_EMERALD2)
		or (mo.type == MT_EMERALD3)
		or (mo.type == MT_EMERALD4)
		or (mo.type == MT_EMERALD5)
		or (mo.type == MT_EMERALD6)
		or (mo.type == MT_EMERALD7)
			table.insert(tokens, mo)
		end
	end

	-- We need to PRECALCULATE this, because it can change when 

	if (#tokens > 0 and #monitors > 0)
		for i = 1,#tokens
			if (#monitors <= 0)
				break
			end

			local swap = (randomizer.rv % #monitors) + 1
			local token = tokens[i]
			local monitor = monitors[swap]

			randomizer.rv = randomizer.scramble(randomizer.rv)

			local newx = monitor.x
			local newy = monitor.y
			local newz = monitor.z
			local newflip = (monitor.flags2 & MF2_OBJECTFLIP)

			if (newflip)
				newz = $1 + monitor.height - token.height - 24*token.scale
			else
				newz = $1 + 24*token.scale
			end

			P_TeleportMove(monitor, token.x, token.y, token.z)
			if (token.flags2 & MF2_OBJECTFLIP)
				monitor.eflags = $1|MFE_VERTICALFLIP
				monitor.flags2 = $1|MF2_OBJECTFLIP
			else
				monitor.eflags = $1 & ~MFE_VERTICALFLIP
				monitor.flags2 = $1 & ~MF2_OBJECTFLIP
			end

			if (token.type == MT_TOKEN)
				randomizer.debugPrint("Found token with ID " .. token.health)

				if (randotokenlist & token.health)
					if (token.tracer and token.tracer.valid)
						P_RemoveMobj(token.tracer)
					end

					P_RemoveMobj(token)
					table.remove(monitors, swap)
					continue
				end
			end

			P_TeleportMove(token, newx, newy, newz)
			if (newflip)
				token.eflags = $1|MFE_VERTICALFLIP
				token.flags2 = $1|MF2_OBJECTFLIP
			else
				token.eflags = $1 & ~MFE_VERTICALFLIP
				token.flags2 = $1 & ~MF2_OBJECTFLIP
			end

			if (token.tracer and token.tracer.valid)
				if (newflip)
					newz = $1 + token.height - token.tracer.height
				end

				P_TeleportMove(token.tracer, newx, newy, newz)
				if (newflip)
					token.tracer.eflags = $1|MFE_VERTICALFLIP
					token.tracer.flags2 = $1|MF2_OBJECTFLIP
				else
					token.tracer.eflags = $1 & ~MFE_VERTICALFLIP
					token.tracer.flags2 = $1 & ~MF2_OBJECTFLIP
				end
			end

			table.remove(monitors, swap)
		end
	end
end

local function removeEmeralds()
	randomizer.debugPrint("Replacing Chaos Emeralds & tokens with monitors")

	local monitortypes = generateMonitorTypesTable()
	local removelater = {}

	for mo in mobjs.iterate()
		if not (mo and mo.valid)
			continue
		end

		if (mo.type == MT_TOKEN)
		or (mo.type == MT_EMERALD1)
		or (mo.type == MT_EMERALD2)
		or (mo.type == MT_EMERALD3)
		or (mo.type == MT_EMERALD4)
		or (mo.type == MT_EMERALD5)
		or (mo.type == MT_EMERALD6)
		or (mo.type == MT_EMERALD7)
			local monitor = P_SpawnMobj(mo.x, mo.y, mo.z, monitortypes[(randomizer.rv % #monitortypes) + 1])
			if (mo.flags2 & MF2_OBJECTFLIP)
				monitor.eflags = $1|MFE_VERTICALFLIP
				monitor.flags2 = $1|MF2_OBJECTFLIP
			else
				monitor.eflags = $1 & ~MFE_VERTICALFLIP
				monitor.flags2 = $1 & ~MF2_OBJECTFLIP
			end

			randomizer.rv = randomizer.scramble(randomizer.rv)

			table.insert(removelater, mo)
		end
	end

	if (#removelater > 0)
		for i = 1,#removelater
			if (removelater[i] and removelater[i].valid)
				P_RemoveMobj(removelater[i])
			end
		end
	end
end

local function shuffleEmblems()
	randomizer.debugPrint("Shuffling Emblems")

	local spots = {}
	local emblems = {}

	for mo in mobjs.iterate()
		if not (mo and mo.valid)
			continue
		end

		if (monitorIsSwappable(mo.info))
		or (mo.type == MT_TOKEN)
		or (mo.type == MT_EMERALD1)
		or (mo.type == MT_EMERALD2)
		or (mo.type == MT_EMERALD3)
		or (mo.type == MT_EMERALD4)
		or (mo.type == MT_EMERALD5)
		or (mo.type == MT_EMERALD6)
		or (mo.type == MT_EMERALD7)
			if ((mo.flags & MF_MONITOR) and (mo.tracer and mo.tracer.valid)) -- 1up overlay
				P_RemoveMobj(mo.tracer)
			end
			table.insert(spots, mo)
		elseif (mo.type == MT_EMBLEM)
			table.insert(emblems, mo)
		end
	end

	if (#emblems > 0 and #spots > 0)
		for i = 1,#emblems
			local swap = (randomizer.rv % #spots) + 1
			local emblem = emblems[i]
			local othermo = spots[swap]

			randomizer.rv = randomizer.scramble(randomizer.rv)

			local newx = othermo.x
			local newy = othermo.y
			local newz = othermo.z
			local newflip = (othermo.flags2 & MF2_OBJECTFLIP)

			if (newflip)
				newz = $1 + othermo.height - emblem.height
				if (othermo.flags & MF_MONITOR)
					newz = $1 - 24*emblem.scale
				end
			else
				if (othermo.flags & MF_MONITOR)
					newz = $1 + 24*emblem.scale
				end
			end

			P_TeleportMove(othermo, emblem.x, emblem.y, emblem.z)
			othermo.eflags = $1 & ~MFE_VERTICALFLIP
			othermo.flags2 = $1 & ~MF2_OBJECTFLIP

			P_TeleportMove(emblem, newx, newy, newz)
			if (newflip)
				emblem.eflags = $1|MFE_VERTICALFLIP
				emblem.flags2 = $1|MF2_OBJECTFLIP
			else
				emblem.eflags = $1 & ~MFE_VERTICALFLIP
				emblem.flags2 = $1 & ~MF2_OBJECTFLIP
			end

			table.remove(spots, swap)
		end
	end
end

addHook("MapLoad", function(map)
	if (randomizer.seed == nil or randomizer.rv == nil)
		return
	end

	randomizer.debugPrint(" == Prize randomization == ")

	if (lastmap != map)
		randomizer.debugPrint("Reset token list")
		randotokenlist = 0
		lastmap = map
	end

	local spots = generateMonitorSpotsTable()

	if (#spots <= 0)
		randomizer.debugPrint("No spots for prizes")
		return
	end

	if (randomizer.cvars.monitors.value == 1)
		replaceMonitors(spots)
	elseif (randomizer.cvars.monitors.value == 2)
		shuffleMonitors(spots)
	else
		randomizer.debugPrint("Monitor randomization skipped")
	end

	if (randomizer.cvars.emeralds.value == 1)
		shuffleEmeralds()
	elseif (randomizer.cvars.emeralds.value == 2)
		removeEmeralds()
	else
		randomizer.debugPrint("Emerald randomization skipped")
	end

	if (randomizer.cvars.emblems.value == 1)
		shuffleEmblems()
	else
		randomizer.debugPrint("Emblem randomization skipped")
	end
end)

addHook("TouchSpecial", function(mo, toucher)
	if (randomizer.seed == nil or randomizer.rv == nil)
		return
	end

	local realbit = mo.health
	mo.health = 0 -- Prevent the game from removing it on its own

	if not (toucher and toucher.valid)
		return
	end

	--[[
	if (!(gametyperules & GTR_SPECIALSTAGES) or modeattacking)
		return
	end
	--]]

	if not (toucher.player.bot)
		randotokenlist = $1 | realbit
		randomizer.debugPrint("Token ID " .. realbit .. " gotten")
	end
end, MT_TOKEN)

addHook("NetVars", function(net)
	randotokenlist = net(randotokenlist) // should only get used in SP, but let's be safe
	lastmap = net(lastmap)
end)
