-- == SRB2Randomizer: Ring randomization ==
-- Places new rings in pre-defined layouts throughout the map

randomizer.cvars.rings = CV_RegisterVar({
	name = "randomizer_rings",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = {
		Off = 0,
		On = 1,
		NeverCeilings = 2,
		AlwaysCeilings = 3,
	},
	defaultvalue = "On",
})

randomizer.newHelpEntry(
	"rings",
	"<integer/string>",
	"Changes ring randomization style.",
	{
		"Changes ring layout randomization style.",
		"All settings (besides \"Off\") will cause some lag when you load a new level (not on reload, though). If this annoys you, then this may be turned off.",
		" This setting also controls Blue Sphere randomization in Special Stages. Options are...",
		" * \130Off (0):\128 Ring layouts will not be randomized.",
		" * \130On (1):\128 Rings will be randomly placed, using predefined patterns. Rings will be placed on ceilings in maps with gravity flipping. (default option)",
		" * \130NeverCeilings (2):\128 Rings will never spawn on the ceiling, even in gravity flip maps. Can increase the ring density on those maps.",
		" * \130AlwaysCeilings (3):\128 Rings can spawn on ceilings in any maps. Thins out the number of collectable rings on most maps.",
	},
	false
)

randomizer.cvars.ringcount = CV_RegisterVar({
	name = "randomizer_ringcount",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = CV_OnOff,
	defaultvalue = "On",
})

randomizer.newHelpEntry(
	"ringcount",
	"<boolean>",
	"Randomize the number of total rings per level.",
	{
		"Decide whenever or not to randomize the number of total rings per level. Options are...",
		" * \130Off (0):\128 The total ring count will try to match the original level's ring count.",
		" * \130On (1):\128 Total ring count can vary from the original level. Makes Perfect Bonus even more likely to be impossible. (default option)",
	},
	false
)

-- Default pattern (just a single ring)
local defaultPattern = {
	[1] = {x = 0, y = 0, z = 0},
}

local PATTERNSUCCESS = 0 -- no problems!
local PATTERNFAILURE_INVALID = 1 -- completely invalid (too many rings, can't fit vertically, etc)
local PATTERNFAILURE_XPOS = 2 -- blocked positive X
local PATTERNFAILURE_XNEG = 4 -- blocked negative X
local PATTERNFAILURE_YPOS = 8 -- blocked positive Y
local PATTERNFAILURE_YNEG = 16 -- blocked negative Y
local PATTERNFAILURE_CENTER = 32 -- blocked center

local function usableSectorSpecial(s)
	local spec = GetSecSpecial(s.special, 1)

	if (spec == 1) -- Damage
	or (spec == 2) -- Water Damage
	or (spec == 3) -- Fire Damage
	or (spec == 4) -- Electric Damage
	or (spec == 5) -- Spike Damage
	or (spec == 6) or (spec == 7) -- Death Pit
	or (spec == 8) -- Instant Kill
	--or (spec == 9) or (spec == 10) -- Ring Drain
	or (spec == 11) -- SS Damage
		return false
	end

	if (GetSecSpecial(s.special, 4) == 2) -- Exit sector
		return false
	end

	return true
end

local function sectorIsSkyFloor(s, flip)
	-- Technically sky floor COULD be a mapper-intended place to put rings,
	-- but this looks nicer anyway.

	if (flip)
		return (s.ceilingpic == "F_SKY1")
	else
		return (s.floorpic == "F_SKY1")
	end
end

local function newSectorInfo(subsec)
	if not (subsec and subsec.valid)
		return nil
	end

	local sec = subsec.sector

	if not (sec and sec.valid)
		return nil
	end

	local sectorinfo = {
		sector = #sec,
		xmin = nil,
		xmax = nil,
		ymin = nil,
		ymax = nil,
	}

	for line in lines.iterate do
		--local line = sec.lines[i]
		if (line.frontsector != sec) and (line.backsector != sec)
			continue
		end

		if (sectorinfo.xmin == nil) or (line.v1.x < sectorinfo.xmin)
			sectorinfo.xmin = line.v1.x
		end
		if (line.v2.x < sectorinfo.xmin)
			sectorinfo.xmin = line.v2.x
		end

		if (sectorinfo.xmax == nil) or (line.v1.x > sectorinfo.xmax)
			sectorinfo.xmax = line.v1.x
		end
		if (line.v2.x > sectorinfo.xmax)
			sectorinfo.xmax = line.v2.x
		end

		if (sectorinfo.ymin == nil) or (line.v1.y < sectorinfo.ymin)
			sectorinfo.ymin = line.v1.y
		end
		if (line.v2.y < sectorinfo.ymin)
			sectorinfo.ymin = line.v2.y
		end

		if (sectorinfo.ymax == nil) or (line.v1.y > sectorinfo.ymax)
			sectorinfo.ymax = line.v1.y
		end
		if (line.v2.y > sectorinfo.ymax)
			sectorinfo.ymax = line.v2.y
		end

		--i = $1+1
	end

	assert(
		(sectorinfo.xmin != nil) and (sectorinfo.xmax != nil)
		and (sectorinfo.ymin != nil) and (sectorinfo.ymax != nil),
		"Bad sector size in newSectorInfo"
	)

	return sectorinfo
end

local function enoughSpaceAboveFloor(requirement, sec, floor, ceiling, reverse)
	local space = 0

	if (reverse == true)
		space = floor - ceiling
	elseif (reverse == false)
		space = ceiling - floor
	end

	if (space < requirement)
		return false
	end

	local newceiling = ceiling

	for rover in sec.ffloors()
		if not (rover.flags & FF_EXISTS)
			continue
		end

		if ((rover.flags & FF_BLOCKPLAYER)
		--or (rover.flags & FF_BLOCKOTHERS)
		or (G_IsSpecialStage() == true and (rover.flags & FF_SWIMMABLE))
		or (usableSectorSpecial(rover.sector) == false)
		or (sectorIsSkyFloor(rover.sector, not reverse) == true))
			if ((rover.topheight <= floor) and (reverse == false))
			or ((rover.bottomheight >= floor) and (reverse == true))
				continue
			end

			if (space < requirement)
				return false
			end

			if ((rover.bottomheight < newceiling) and (reverse == false))
				newceiling = rover.bottomheight
			elseif ((rover.topheight > newceiling) and (reverse == true))
				newceiling = rover.topheight
			end
		end
	end

	if (reverse == true)
		space = floor - newceiling
	elseif (reverse == false)
		space = newceiling - floor
	end

	if (space < requirement)
		return false
	end

	return true
end

local function newSpot(x, y, z, slope, flip, springtype, angle)
	local spotinfo = {}

	spotinfo.x = x
	spotinfo.y = y
	spotinfo.z = z
	spotinfo.slope = slope
	spotinfo.flip = flip
	spotinfo.springtype = springtype
	spotinfo.angle = angle

	return spotinfo
end

local function goodz(za, zb, slopea, slopeb)
	if (slopea and slopea.valid)
		if (slopeb and slopeb.valid) and (slopeb.o == slopea.o) 
			return true
		end
	else
		if (za == zb)
			return true
		end
	end

	return false
end

local function validPattern(pattern, numrings, x, y, z, slope, angle, ceiling)
	local warningflags = PATTERNSUCCESS

	if (#pattern > numrings)
		-- Not enough rings to place this pattern in the first place
		return PATTERNFAILURE_INVALID
	end

	--[[
	local patternheight = 0
	for i = 1,#pattern
		if (pattern[i].z > patternheight)
			patternheight = pattern[i].z
		end
	end

	patternheight = (($1 + 24) * FRACUNIT) + mobjinfo[MT_RING].height + (24*FRACUNIT) -- add floating height, thing height, and player jump height

	-- We're not too fussed about this, so just check the center once at the start.
	local startsec = R_PointInSubsector(x, y).sector
	if (enoughSpaceAboveFloor(patternheight, startsec, startsec.floorheight, startsec.ceilingheight, true) == false)
		-- Swap the pattern out immediately if it doesn't pass this extremely rudimentary easy test
		return PATTERNFAILURE_INVALID
	end
	--]]

	-- Make a square bounding box for the pattern, make sure each corner is at a similar height
	local x1 = nil
	local x2 = nil
	local y1 = nil
	local y2 = nil

	for i = 1,#pattern
		local thisx = (pattern[i].x * cos(angle)) - (pattern[i].y * sin(angle))
		local thisy = (pattern[i].y * cos(angle)) + (pattern[i].x * sin(angle))
		
		if (x1 == nil) or (thisx < x1) then x1 = thisx end
		if (x2 == nil) or (thisx > x2) then x2 = thisx end
		if (y1 == nil) or (thisy < y1) then y1 = thisy end
		if (y2 == nil) or (thisy > y2) then y2 = thisy end
	end

	-- Buffer with a little extra space to keep it out of the walls
	x1 = $1 - (64 * FRACUNIT)
	x2 = $1 + (64 * FRACUNIT)
	y1 = $1 - (64 * FRACUNIT)
	y2 = $1 + (64 * FRACUNIT)

	for i = -1,3
		local flag = 0
		local tryx = x
		local tryy = y

		if (i == -1)
			flag = $1|PATTERNFAILURE_CENTER
		else
			if (i & 1)
				flag = $1|PATTERNFAILURE_XPOS
				tryx = $1 + x2
			else
				flag = $1|PATTERNFAILURE_XNEG
				tryx = $1 + x1
			end

			if (i > 1)
				flag = $1|PATTERNFAILURE_YPOS
				tryy = $1 + y2
			else
				flag = $1|PATTERNFAILURE_YNEG
				tryy = $1 + y1
			end
		end

		local validz = false

		local subsec = R_PointInSubsectorOrNil(tryx, tryy)
		local sec = nil

		if (subsec and subsec.valid)
			sec = subsec.sector
		end

		if (sec and sec.valid)
			-- Make sure to check those FOFs!!
			if (ceiling)
				if (usableSectorSpecial(sec) == true
				and sectorIsSkyFloor(sec, true) == false
				and goodz(z, sec.ceilingheight, slope, sec.c_slope))
					validz = true
				else
					for rover in sec.ffloors()
						if (not (rover.flags & FF_EXISTS))
						or (rover.flags & FF_BUSTUP)
						or (rover.flags & FF_REVERSEPLATFORM)
							continue
						end

						if (usableSectorSpecial(rover.sector) == false)
						or (sectorIsSkyFloor(rover.sector, true) == true)
							continue
						end

						if (rover.flags & FF_BLOCKPLAYER)
						--or (rover.flags & FF_BLOCKOTHERS)
						or ((rover.flags & FF_SWIMMABLE) and (rover.flags & FF_GOOWATER) and G_IsSpecialStage() == false)
							if (goodz(z, rover.bottomheight, slope, rover.b_slope) == true)
								validz = true
								break
							end
						end
					end
				end
			else
				if (usableSectorSpecial(sec) == true
				and sectorIsSkyFloor(sec, false) == false
				and goodz(z, sec.floorheight, slope, sec.f_slope))
					validz = true
				else
					for rover in sec.ffloors()
						if (not (rover.flags & FF_EXISTS))
						or (rover.flags & FF_BUSTUP)
						or (rover.flags & FF_REVERSEPLATFORM)
							continue
						end

						if (usableSectorSpecial(rover.sector) == false)
						or (sectorIsSkyFloor(rover.sector, false) == true)
							continue
						end

						if (rover.flags & FF_BLOCKPLAYER)
						--or (rover.flags & FF_BLOCKOTHERS)
						or ((rover.flags & FF_SWIMMABLE) and (rover.flags & FF_GOOWATER) and G_IsSpecialStage() == false)
							if (goodz(z, rover.topheight, slope, rover.t_slope))
								validz = true
								break
							end
						end
					end
				end
			end
		end

		if (validz == false)
			warningflags = $1|flag
		end
	end

	return warningflags
end

local function spawnRingPattern(pattern, x, y, z, slope, ang, flip, type, xymul, zmul)
	for i = 1,#pattern
		local ringpos = pattern[i]
		local newx = x + (ringpos.x * FixedMul(xymul, cos(ang))) - (ringpos.y * FixedMul(xymul, sin(ang)))
		local newy = y + (ringpos.y * FixedMul(xymul, cos(ang))) + (ringpos.x * FixedMul(xymul, sin(ang)))
		local newz = z

		if (flip)
			newz = $1 - mobjinfo[type].height - FixedMul((24+ringpos.z)*FRACUNIT, zmul)
		else
			newz = $1 + FixedMul((24+ringpos.z)*FRACUNIT, zmul)
		end

		if (slope and slope.valid)
			newz = $1 + (P_GetZAt(slope, newx, newy) - z)
		end

		local newring = P_SpawnMobj(newx, newy, newz, type)

		if (flip)
			newring.eflags = $1|MFE_VERTICALFLIP
			newring.flags2 = $1|MF2_OBJECTFLIP
		end
	end
end

local spots = {}
local lastmap = nil

local function generateRingSpots()
	randomizer.debugPrint("Generating new ring spots")
	spots = {}

	-- Figure out if it's OK to put rings on the ceilings
	local allowceilings = false

	if (randomizer.cvars.rings.value == 2)
		allowceilings = false
	elseif (randomizer.cvars.rings.value == 3)
		allowceilings = true
	else
		for l in lines.iterate do
			if (l.special == 433) -- Enable/Disable Gravity Flip 
				allowceilings = true
				break
			end
		end
	end

	-- Figure out size of the map, so we can reduce how often we need to calculate things
	local mapsize = 0
	local mapminx = 0
	local mapmaxx = 0
	local mapminy = 0
	local mapmaxy = 0

	for i = 0,(#vertexes)-1
		local vertex = vertexes[i]

		if not (vertex and vertex.valid)
			continue
		end

		local vx = vertex.x / FRACUNIT
		local vy = vertex.y / FRACUNIT

		if (vx < mapminx)
			mapminx = vx
		end
		if (vx > mapmaxx)
			mapmaxx = vx
		end
		if (vy < mapminy)
			mapminy = vy
		end
		if (vy > mapmaxy)
			mapmaxy = vy
		end
	end

	mapsize = min(mapmaxx - mapminx, mapmaxy - mapminy)
	assert(mapsize > 0, "Map has size of 0")

	randomizer.debugPrint("Map size: " .. mapsize)

	local spacing = mapsize / 20
	if (spacing < 256)
		spacing = 256
	end

	randomizer.debugPrint("Pattern spacing: " .. spacing)

	randomizer.debugPrint("Adding generic spots...")

	local checkx = mapminx + (spacing >> 1) - spacing
	while (checkx < mapmaxx - (spacing >> 1))
		checkx = $1 + spacing

		local checky = mapminy + (spacing >> 1) - spacing
		while (checky < mapmaxy - (spacing >> 1))
			checky = $1 + spacing

			local fixedx = checkx * FRACUNIT
			local fixedy = checky * FRACUNIT

			local sinfo = newSectorInfo(R_PointInSubsectorOrNil(fixedx, fixedy))

			if (sinfo == nil)
				continue
			end

			if (fixedx < sinfo.xmin) or (fixedx > sinfo.xmax)
			or (fixedy < sinfo.ymin) or (fixedy > sinfo.ymax)
				continue
			end

			local sec = sectors[sinfo.sector]

			-- Calculate spots
			if (usableSectorSpecial(sec))
				if (sectorIsSkyFloor(sec, false) == false)
				and (enoughSpaceAboveFloor(128*FRACUNIT, sec, sec.floorheight, sec.ceilingheight, false))
					table.insert(spots, newSpot(fixedx, fixedy, sec.floorheight, sec.f_slope, false))
				end

				if (allowceilings)
				and (sectorIsSkyFloor(sec, true) == false)
				and (enoughSpaceAboveFloor(128*FRACUNIT, sec, sec.ceilingheight, sec.floorheight, true))
					table.insert(spots, newSpot(fixedx, fixedy, sec.ceilingheight, sec.c_slope, true))
				end
			end

			for rover in sec.ffloors()
				if not (rover.flags & FF_EXISTS)
					continue
				end

				if (usableSectorSpecial(rover.sector) == false)
					continue
				end

				if not ((rover.flags & FF_BLOCKPLAYER)
				--or (rover.flags & FF_BLOCKOTHERS)
				or ((rover.flags & FF_SWIMMABLE) and (rover.flags & FF_GOOWATER) and G_IsSpecialStage() == false))
					continue
				end
				
				if (rover.flags & FF_BUSTUP)
				or ((rover.flags & FF_CRUMBLE) and (rover.flags & FF_NORETURN))
					continue
				end

				if (sectorIsSkyFloor(rover.sector, true) == true)
				and (enoughSpaceAboveFloor(128*FRACUNIT, sec, rover.topheight, sec.ceilingheight, false))
					table.insert(spots, newSpot(fixedx, fixedy, rover.topheight, rover.t_slope, false))
				end

				if (allowceilings)
				and (sectorIsSkyFloor(rover.sector, false) == true)
				and (enoughSpaceAboveFloor(128*FRACUNIT, sec, rover.bottomheight, sec.floorheight, true))
					table.insert(spots, newSpot(fixedx, fixedy, rover.bottomheight, rover.b_slope, true))
				end
			end
		end
	end

	randomizer.debugPrint("Adding spring spots...")

	for mo in mobjs.iterate()
		if not (mo and mo.valid)
			continue
		end

		if ((mo.flags & MF_SPRING) or (mo.type == MT_STEAM))
			local moinfo = mobjinfo[mo.type]

			if (moinfo.mass == 0) or (moinfo.painchance == -1)
				continue
			end

			local vflip = (mo.eflags & MFE_VERTICALFLIP)
			local maxheight = FixedMul(FixedMul(mo.info.mass, mo.info.mass), mo.scale)

			local floor = mo.z
			local ceil = mo.subsector.sector.ceilingheight

			if (moinfo.mass < 0)
				vflip = !vflip
			end

			if (vflip)
				floor = mo.z + mo.height
				ceil = mo.subsector.sector.floorheight
			end

			-- Ensure there's enough space
			if (enoughSpaceAboveFloor(maxheight, mo.subsector.sector, floor, ceil, vflip))
				table.insert(spots, newSpot(mo.x, mo.y, mo.z, nil, vflip, mo.type, mo.angle))
			end
		end
	end

	randomizer.debugPrint("Finished generating spots!")
end

local function createRings()
	-- Spawn coins instead in Mario Mode
	local ringtype = MT_RING
	if (mariomode)
		ringtype = MT_COIN
	end

	-- Count number of rings, then remove them
	local oldrings = {}
	local oldringcount = 0
	local ringcount = 0

	local oldspheres = {}
	local oldspherecount = 0
	local spherecount = 0
	local requiredspheres = 0

	for mo in mobjs.iterate()
		if not (mo and mo.valid)
			continue
		end

		if (mo.type == MT_RING) or (mo.type == MT_COIN)
			table.insert(oldrings, mo)
		elseif (mo.type == MT_BLUESPHERE)
			table.insert(oldspheres, mo)
		end
	end

	oldringcount = #oldrings
	ringcount = oldringcount
	oldspherecount = #oldspheres
	spherecount = oldspherecount

	if (G_IsSpecialStage())
		requiredspheres = 1

		for sec in sectors.iterate do
			if (GetSecSpecial(sec.special, 2) == 10)
				-- special stage parameters
				requiredspheres = sec.ceilingheight / FRACUNIT
			end
		end

		-- Ensure there are enough spheres by trying to spawn extra
		local prevspherecount = spherecount
		spherecount = $1 + ($1 / 10) -- +10% more spheres
		randomizer.debugPrint("Boosted sphere count from " .. prevspherecount .. " to " .. spherecount)
	end

	if (ringcount <= 0 and spherecount <= 0)
		-- Don't need to do anything in a map with 0 rings.
		randomizer.debugPrint("No rings or spheres, skipping")
		return
	end

	if (randomizer.cvars.ringcount.value == 1)
		local maxringchange = ringcount >> 1

		local ringoffset = (randomizer.rv % (maxringchange + 1))
		randomizer.rv = randomizer.scramble(randomizer.rv)

		local subtract = (randomizer.rv & 1)
		randomizer.rv = randomizer.scramble(randomizer.rv)

		if (subtract)
			ringcount = $1 - ringoffset
		else
			ringcount = $1 + ringoffset
		end

		if (ringcount < 0)
			ringcount = 0
		end

		randomizer.debugPrint("Final intended ring count: " .. ringcount)
	end

	for i = 1,#oldrings
		if (oldrings[i] and oldrings[i].valid)
			oldrings[i].flags2 = $1|MF2_DONTRESPAWN
			P_RemoveMobj(oldrings[i])
		end
	end

	for i = 1,#oldspheres
		if (oldspheres[i] and oldspheres[i].valid)
			oldspheres[i].flags2 = $1|MF2_DONTRESPAWN
			P_RemoveMobj(oldspheres[i])
		end
	end

	-- Only generate a new ring spots table IF we need to.
	if (lastmap != gamemap)
		generateRingSpots()
		lastmap = gamemap
	else
		randomizer.debugPrint("Using pre-compiled spots table.")
	end

	local tempspots = randomizer.simpleTableCopy(spots)
	local leftovers = {} -- Used as fallbacks for any 

	while (ringcount > 0 or spherecount > 0)
		if (#tempspots <= 0)
			-- No more spots to use for ring placement
			randomizer.debugPrint("No more valid spots!")
			break
		end

		local spotindex = (randomizer.rv % #tempspots) + 1
		local usespot = tempspots[spotindex]
		randomizer.rv = randomizer.scramble(randomizer.rv)

		local originx = usespot.x
		local originy = usespot.y
		local originz = usespot.z
		local onceiling = usespot.flip

		local numangles = 8
		local anglenum = 0
		local offsetangle = 0
		local rotateattempts = numangles-1

		local temppatterns

		if (usespot.springtype != nil)
			if (mobjinfo[usespot.springtype].damage)
				offsetangle = usespot.angle
				temppatterns = randomizer.simpleTableCopy(randomizer.randomDSpringPatterns)
			else
				temppatterns = randomizer.simpleTableCopy(randomizer.randomVSpringPatterns)
			end
		else
			anglenum = (randomizer.rv % numangles)
			randomizer.rv = randomizer.scramble(randomizer.rv)
			offsetangle = FixedAngle(anglenum * ((360*FRACUNIT) / numangles))
			temppatterns = randomizer.simpleTableCopy(randomizer.randomRingPatterns)
		end

		local startx = originx
		local starty = originy

		assert(#temppatterns > 0, "No ring patterns found")

		local patternindex = (randomizer.rv % #temppatterns) + 1
		randomizer.rv = randomizer.scramble(randomizer.rv)

		local pattern = temppatterns[patternindex]
		table.remove(temppatterns, patternindex)

		local count = ringcount

		if (spherecount > 0)
			-- Prioritize placing spheres over rings
			count = spherecount
		end

		local isPatternValid = validPattern(pattern, count, originx, originy, originz, usespot.slope, offsetangle, onceiling)
		local prevValid = isPatternValid

		if (usespot.springtype != nil) -- Can't adjust position for spring spots, but they're most likely OK anyway
			isPatternValid = PATTERNSUCCESS
			prevValid = isPatternValid

			while (#pattern > count)
				if (#temppatterns <= 0)
					-- Out of attempts, just give up & place a single ring.
					pattern = defaultPattern
					break
				end

				patternindex = (randomizer.rv % #temppatterns) + 1
				randomizer.rv = randomizer.scramble(randomizer.rv)

				pattern = temppatterns[patternindex]
				table.remove(temppatterns, patternindex)
			end

			if (#pattern > count)
				isPatternValid = PATTERNFAILURE_INVALID
			end
		else
			while (isPatternValid != PATTERNSUCCESS)
				local newpattern = false -- Decide whenever or not to try out another pattern
				local rotated = false

				if (isPatternValid & PATTERNFAILURE_INVALID) 
					-- Pattern had too many rings
					newpattern = true
				end

				if (isPatternValid & PATTERNFAILURE_CENTER)
					-- The current pattern's center was blocked entirely after nudging
					newpattern = true
				end

				if ((isPatternValid & PATTERNFAILURE_XPOS) and (isPatternValid & PATTERNFAILURE_XNEG))
				or ((isPatternValid & PATTERNFAILURE_XPOS) and (prevValid & PATTERNFAILURE_XNEG))
				or ((isPatternValid & PATTERNFAILURE_XNEG) and (prevValid & PATTERNFAILURE_XPOS))
					if (rotateattempts)
						-- Maybe the pattern can fit if we rotate it?
						if not (rotated)
							anglenum = ($1+1) % numangles
							offsetangle = FixedAngle(anglenum * ((360*FRACUNIT) / numangles))
							rotateattempts = $1-1
							rotated = true
						end
					else
						-- The current pattern will never have enough space on the X axis
						newpattern = true
					end
				end

				if ((isPatternValid & PATTERNFAILURE_YPOS) and (isPatternValid & PATTERNFAILURE_YNEG))
				or ((isPatternValid & PATTERNFAILURE_YPOS) and (prevValid & PATTERNFAILURE_YNEG))
				or ((isPatternValid & PATTERNFAILURE_YNEG) and (prevValid & PATTERNFAILURE_YPOS))
					if (rotateattempts)
						-- Maybe the pattern can fit if we rotate it?
						if not (rotated)
							anglenum = ($1+1) % numangles
							offsetangle = FixedAngle(anglenum * ((360*FRACUNIT) / numangles))
							rotateattempts = $1-1
							rotated = true
						end
					else
						-- The current pattern will never have enough space on the Y axis
						newpattern = true
					end
				end

				if (newpattern)
					if (#temppatterns <= 0)
						-- Out of attempts, just give up & place a single ring.
						pattern = defaultPattern
						break
					end

					originx = startx
					originy = starty
					rotateattempts = numangles-1

					patternindex = (randomizer.rv % #temppatterns) + 1
					randomizer.rv = randomizer.scramble(randomizer.rv)

					pattern = temppatterns[patternindex]
					table.remove(temppatterns, patternindex)
				else
					-- Try to nudge it away from the failing edge
					if (isPatternValid & PATTERNFAILURE_XPOS)
						originx = $1 - (64*FRACUNIT)
					elseif (isPatternValid & PATTERNFAILURE_XNEG)
						originx = $1 + (64*FRACUNIT)
					end

					if (isPatternValid & PATTERNFAILURE_YPOS)
						originy = $1 - (64*FRACUNIT)
					elseif (isPatternValid & PATTERNFAILURE_YNEG)
						originy = $1 + (64*FRACUNIT)
					end
				end

				prevValid = isPatternValid
				isPatternValid = validPattern(pattern, ringcount, originx, originy, originz, usespot.slope, offsetangle, onceiling)
				if (newpattern)
					prevValid = isPatternValid
				end
			end
		end

		-- EXPERIMENTAL: Try another spot if it could not find anything usable here,
		-- instead of just defaulting to a single ring.
		if (isPatternValid != PATTERNSUCCESS)
			randomizer.rv = randomizer.scramble(randomizer.rv)
			if (usespot.springtype == nil)
				table.insert(leftovers, usespot)
			end
			table.remove(tempspots, spotindex)
			continue
		end

		local xymul = FRACUNIT
		local zmul = FRACUNIT

		if (usespot.springtype != nil)
			zmul = FixedDiv(mobjinfo[usespot.springtype].mass, mobjinfo[MT_YELLOWSPRING].mass)
			zmul = FixedMul($1, $1)

			if (mobjinfo[usespot.springtype].damage)
				xymul = FixedDiv(mobjinfo[usespot.springtype].damage, mobjinfo[MT_YELLOWDIAG].damage)
				xymul = FixedMul($1, $1)
			end
		end

		if (spherecount > 0)
			spawnRingPattern(pattern, originx, originy, originz, usespot.slope, offsetangle, onceiling, MT_BLUESPHERE, xymul, zmul)
			spherecount = $1 - #pattern
			randomizer.debugPrint("Spawned blue sphere pattern @ " .. (originx / FRACUNIT) .. ", " .. (originy / FRACUNIT) )
		else
			spawnRingPattern(pattern, originx, originy, originz, usespot.slope, offsetangle, onceiling, ringtype, xymul, zmul)
			ringcount = $1 - #pattern
			randomizer.debugPrint("Spawned ring pattern @ " .. (originx / FRACUNIT) .. ", " .. (originy / FRACUNIT) )
		end

		randomizer.rv = randomizer.scramble(randomizer.rv)
		table.remove(tempspots, spotindex)
	end

	if (spherecount > 0 or ringcount > 0)
		-- Restore pre-maturely removed spots
		for i = 1,#leftovers
			table.insert(tempspots, leftovers[i])
		end

		while ((spherecount > 0 or ringcount > 0) and #tempspots > 0)
			-- Still rings left? Place single rings in any leftover spots.

			local spotindex = (randomizer.rv % #tempspots) + 1
			local usespot = tempspots[spotindex]
			randomizer.rv = randomizer.scramble(randomizer.rv)

			if (spherecount > 0)
				spawnRingPattern(defaultPattern, usespot.x, usespot.y, usespot.z, nil, 0, usespot.flip, MT_BLUESPHERE, FRACUNIT, FRACUNIT)
				spherecount = $1 - #defaultPattern
				randomizer.debugPrint("Spawned single blue sphere @ " .. (usespot.x / FRACUNIT) .. ", " .. (usespot.y / FRACUNIT) )
			else
				spawnRingPattern(defaultPattern, usespot.x, usespot.y, usespot.z, nil, 0, usespot.flip, ringtype, FRACUNIT, FRACUNIT)
				ringcount = $1 - #defaultPattern
				randomizer.debugPrint("Spawned single ring @ " .. (usespot.x / FRACUNIT) .. ", " .. (usespot.y / FRACUNIT) )
			end

			table.remove(tempspots, spotindex)
		end

		if (ringcount > 0)
			-- *STILL* rings left? Welp, that's all I can do.
			randomizer.debugPrint("Could not place all rings! (" .. ringcount .. " left)")
		end

		if (spherecount > 0)
			-- Ditto, for spheres.
			randomizer.debugPrint("Could not place all blue spheres! (" .. spherecount .. " left)")
		end
	end

	-- Fail-safe:
	-- Spheres and rings are SUPER important in Special Stages!
	-- Quietly spawn them on Player 1 if any are missing.
	if (G_IsSpecialStage())
		local spheresnotspawned = oldspherecount - spherecount
		local spherestogive = requiredspheres - spheresnotspawned

		ringcount = max(0, ringcount)
		spherestogive = max(0, spherestogive)

		if (ringcount > 0 or spherestogive > 0)
			randomizer.debugPrint("Couldn't create enough items for the Special Stage!")
			randomizer.debugPrint("Giving players " .. ringcount .. " rings and " .. spherestogive .. " spheres...")
		end

		for player in players.iterate do
			if (player.mo and player.mo.valid)
			and (player.spectator == false)
				player.rings = $1 + oldringcount -- HACK to make perfect bonus still possible
				player.rings = $1 + max(0, ringcount)
				player.spheres = $1 + max(0, spherestogive)
				ringcount = 0
				spherestogive = 0
				break
			end
		end
	end
end

-- HACK: using iterators during MapLoad can give me very rare invalid userdata
-- so just delay it to the first thinkframe tic.
--local createRingsDelayed = nil 

addHook("MapLoad", function(map)
	if (randomizer.seed == nil or randomizer.rv == nil)
		return
	end

	randomizer.debugPrint(" == Ring & sphere randomization == ")

	if (randomizer.cvars.rings.value == 0)
		randomizer.debugPrint("Turned off, skipping")
		return
	end

	if (maptol & TOL_NIGHTS)
		-- NiGHTS doesn't work with this ring randomization
		randomizer.debugPrint("NiGHTS mode, skipping")
		return
	end

	--createRingsDelayed = map
	createRings(map)
end)

--[[
addHook("ThinkFrame", function()
	if (createRingsDelayed == nil)
		return
	end

	createRings(createRingsDelayed)
	createRingsDelayed = nil
end)
--]]
