-- == SRB2Randomizer: Sky randomization ==
-- Changes all sky textures & weather effects around

randomizer.cvars.skies = CV_RegisterVar({
	name = "randomizer_skies",
	flags = CV_SHOWMODIF,
	PossibleValue = {
		Off = 0,
		KeepSkyboxes = 1,
		NoSkyboxes = 2,
		-- Randomize sky textures. Removes the skybox.
		-- For level packs that have lots of skyboxes that cover the texture.
		RandomSkybox = 3,
		-- Randomize sky textures and skybox points.
		-- For max insanity, or if you like looking at HOMs.
	},
	defaultvalue = "KeepSkyboxes",
})

randomizer.newHelpEntry(
	"skies",
	"<integer/string>",
	"Changes sky randomization style.",
	{
		"Changes sky texture & skybox randomization style. Options are...",
		" * \130Off (0):\128 Normal sky textures & skyboxes.",
		" * \130KeepSkyboxes (1):\128 Randomize sky textures, and keeps the existing skybox. (default option)",
		" * \130NoSkyboxes (2):\128 Randomize sky textures, and removes the skybox. For level packs that have lots of skyboxes that cover the texture.",
		" * \130RandomSkybox (3):\128 Randomize sky textures and skybox points. For max insanity, or if you like looking at HOMs.",
	},
	true
)

randomizer.cvars.weather = CV_RegisterVar({
	name = "randomizer_weather",
	flags = CV_SHOWMODIF,
	PossibleValue = CV_OnOff,
	defaultvalue = "On",
})

randomizer.newHelpEntry(
	"weather",
	"<boolean>",
	"Decides whenever or not to randomize weather.",
	{
		"Decides whenever or not to randomize the map's weather effects. Options are...",
		" * \130Off (0):\128 Use the level's existing weather.",
		" * \130On (1):\128 Randomize each level's weather. (default option)",
	},
	true
)

randomizer.cvars.light = CV_RegisterVar({
	name = "randomizer_light",
	flags = CV_SHOWMODIF,
	PossibleValue = {
		Off = 0,
		On = 1,
		Random = 2,
	},
	defaultvalue = "On",
})

randomizer.newHelpEntry(
	"light",
	"<integer/string>",
	"Changes light level randomization style.",
	{
		"Changes light level randomization style. Options are...",
		" * \130Off (0):\128 Lighting is kept the same.",
		" * \130On (1):\128 Randomize a light offset per level, keeping the general look of the level. (default option)",
		" * \130Random (2):\128 Randomize light levels, with no logic.",
	},
	true
)

-- Pretty much just a copy of https://wiki.srb2.org/wiki/Flats_and_textures/Skies
local skylist = {
	1,
	2,
	3,
	4,
	5,
	6,
	7,
	8,
	9,
	10,
	11,
	12,
	17,
	18,
	19,
	20,
	21,
	22,
	29,
	30,
	31,
	33,
	34,
	35,
	40,
	50,
	51,
	52,
	53,
	54,
	55,
	56,
	57,
	58,
	59,
	64,
	66,
	68,
	98,
	99,
	100,
	103,
	107,
	117,
	127,
	132, -- 133-135 are animation frames
	159,
	300,
	301,
	302,
	1024,
	1025,
	1026,
	1027,
	1028,
}

local savedskynum = nil

local function setRandomizedSky()
	local v = randomizer.seed + gamemap + levelskynum
	v = randomizer.scramble(v)

	savedskynum = skylist[(v % #skylist) + 1]
	P_SetupLevelSky(savedskynum)

	if (randomizer.cvars.skies.value == 2)
		P_SetSkyboxMobj(nil, false)
		P_SetSkyboxMobj(nil, true)
	end
end

local function addSky(num)
	for i = 1,#skylist
		if (num == skylist[i])
			return
		end
	end

	randomizer.debugPrint("Added sky no. " .. num)
	table.insert(skylist, num)
end

local function scrapeForSkies()
	for i = 1,1035
		if (mapheaderinfo[i] == nil)
			continue
		end

		addSky(mapheaderinfo[i].skynum)
	end
end

-- Doesn't NEED to be a list, but I want a bit of weighted probability
local preciplist = {
	PRECIP_NONE,
	PRECIP_NONE,
	PRECIP_NONE,
	PRECIP_NONE,
	PRECIP_NONE,
	PRECIP_RAIN,
	PRECIP_RAIN,
	PRECIP_STORM,
	PRECIP_STORM_NORAIN,
	PRECIP_STORM_NOSTRIKES,
	PRECIP_SNOW,
	PRECIP_SNOW,
	PRECIP_SNOW,
}

local savedweather = nil

local function setRandomizedWeather()
	local v = randomizer.seed + gamemap
	v = randomizer.scramble(v)
	savedweather = preciplist[(v % #preciplist) + 1]
	P_SwitchWeather(savedweather)
end

local function newLightOffset(rv)
	local swthreshold = 8
	local maxoffset = 16 + (swthreshold-1)
	return -maxoffset + (rv % ((2*maxoffset)+1))
end

local function randomizeLight()
	randomizer.debugPrint(" == Light level randomization == ")

	local v = randomizer.seed + gamemap

	local lightoffset = newLightOffset(v)
	v = randomizer.scramble(v)

	for sec in sectors.iterate do
		if (lightoffset != 0)
			local offsetmul = FRACUNIT
			
			if (lightoffset > 0)
				offsetmul = (128 * FRACUNIT / max(sec.lightlevel, 1)) -- Darker sectors are affected more by brightening
			else
				offsetmul = (sec.lightlevel * FRACUNIT / 128) -- Brighter sectors are affected more by darkening
			end

			sec.lightlevel = $1 + (FixedMul(lightoffset * FRACUNIT, offsetmul) / FRACUNIT)
		end

		if (randomizer.cvars.light.value == 2)
			lightoffset = newLightOffset(v)
			v = randomizer.scramble(v)
		end
	end
end

local lastmap = nil
addHook("MapChange", function(map)
	savedskynum = nil
	savedweather = nil

	if (map != lastmap)
		scrapeForSkies()
	end

	lastmap = map
end)

addHook("ThinkFrame", function()
	if (randomizer.seed == nil or randomizer.rv == nil)
		return
	end

	if (randomizer.cvars.skies.value != 0)
	and (savedskynum == nil or levelskynum != savedskynum)
		setRandomizedSky()
	end

	if (randomizer.cvars.weather.value != 0)
	and (savedweather == nil or curWeather != savedweather)
		setRandomizedWeather()
	end
end)

addHook("MapLoad", function()
	if (randomizer.seed == nil or randomizer.rv == nil)
		return
	end

	if (randomizer.cvars.light.value != 0)
		randomizeLight()
	end

	if (randomizer.cvars.skies.value != 3)
		return
	end

	randomizer.debugPrint(" == Skybox point randomization == ")

	local allmos = {}
	for mo in mobjs.iterate()
		table.insert(allmos, mo)
	end

	P_SetSkyboxMobj(allmos[(randomizer.seed % #allmos) + 1], false)
	P_SetSkyboxMobj(allmos[(randomizer.scramble(randomizer.seed) % #allmos) + 1], true)
end)

addHook("NetVars", function(net)
	savedskynum = net(savedskynum)
	savedweather = net(savedweather)
	lastmap = net(lastmap)

	P_SetupLevelSky(savedskynum)
	P_SwitchWeather(savedweather)
end)
